from collections.abc import Sequence
from typing import Unpack

import loguru

from ._icecream import init_icecream
from ._init_loguru import init_loguru


def init_logging(
    *,
    handlers: Sequence["loguru.HandlerConfig"] | None = None,
    levels: Sequence["loguru.LevelConfig"] | None = None,
    **kwargs: Unpack["loguru.BasicHandlerConfig"],
) -> None:
    init_loguru(handlers=handlers, levels=levels, **kwargs)
    init_icecream()



from collections.abc import Callable, Mapping
from pathlib import Path
from typing import Literal

import autoregistry
from rich.style import Style
from rich.text import Text

_func = autoregistry.Registry(prefix="_func_")


def func(obj: Callable, *, style: Literal["short", "long"] = "short") -> Text:
    return _func[style](obj)


@_func
def _func_short(obj: Callable) -> Text:
    text = Text()
    file: Path = Path(obj.__code__.co_filename)
    if file.exists():
        text.append(
            f"{obj.__name__}()",
            style=Style(link=f"{file.as_uri()}#{obj.__code__.co_firstlineno}"),
        )
    else:
        text.append(f"{obj.__name__}()")
    return text


@_func
def _func_long(obj: Callable) -> Text:
    text = Text()
    file: Path = Path(obj.__code__.co_filename)
    if file.exists():
        text.append(obj.__module__, style=Style(link=file.as_uri()))
        text.append(".")
        text.append(
            f"{obj.__qualname__}(...)",
            style=Style(link=f"{file.as_uri()}#{obj.__code__.co_firstlineno}"),
        )
    else:
        text.append(f"{obj.__module__}.{obj.__qualname__}(...)")
    return text


_pretty_func = func


def call(func: Callable, args: tuple, kwargs: Mapping) -> Text:  # noqa: ARG001
    # TODO: add `args` and `kwargs`
    return _pretty_func(func)



import functools
from collections.abc import Callable, Sequence

from liblaf.grapes import pretty

from . import callback
from ._base import Callback, TimerRecords


class TimedFunction[**P, T]:
    timing: TimerRecords
    _func: Callable[P, T]

    def __init__(
        self,
        func: Callable[P, T],
        /,
        name: str | None = None,
        timers: Sequence[str] = ("perf",),
        callback_start: Callback | None = None,
        callback_stop: Callback | None = None,
        callback_finish: Callback | None = None,
    ) -> None:
        if name is None:
            name = pretty.func(func).plain or "Function"
        if callback_stop is None:
            callback_stop = callback.log_record(depth=3)
        if callback_finish is None:
            callback_finish = callback.log_summary(depth=2)
        self.timing = TimerRecords(
            name=name,
            timers=timers,
            callback_start=callback_start,
            callback_stop=callback_stop,
            callback_finish=callback_finish,
        )
        self._func = func
        functools.update_wrapper(self, func)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        self.timing.start()
        result: T = self._func(*args, **kwargs)
        self.timing.stop()
        return result



from ._rich import (
    LevelColumn,
    LocationColumn,
    LoguruRichHandler,
    MessageColumn,
    RichLoggingColumn,
    TimeColumn,
)

__all__ = [
    "LevelColumn",
    "LocationColumn",
    "LoguruRichHandler",
    "MessageColumn",
    "RichLoggingColumn",
    "TimeColumn",
]



import functools
from typing import IO, Literal

import rich
from environs import env
from rich.console import Console
from rich.style import Style
from rich.theme import Theme

from liblaf.grapes import path
from liblaf.grapes.typed import PathLike


def theme() -> Theme:
    return Theme(
        {
            "logging.level.notset": Style(dim=True),
            "logging.level.trace": Style(color="cyan", bold=True),
            "logging.level.debug": Style(color="blue", bold=True),
            "logging.level.icecream": Style(color="magenta", bold=True),
            "logging.level.info": Style(bold=True),
            "logging.level.success": Style(color="green", bold=True),
            "logging.level.warning": Style(color="yellow", bold=True),
            "logging.level.error": Style(color="red", bold=True),
            "logging.level.critical": Style(color="red", bold=True, reverse=True),
        },
        inherit=True,
    )


@functools.cache
def get_console(
    file: Literal["stdout", "stderr"] | IO | PathLike = "stdout", **kwargs
) -> Console:
    match file:
        case "stdout":
            rich.reconfigure(force_terminal=force_terminal(), theme=theme(), **kwargs)
            return rich.get_console()
        case "stderr":
            return Console(
                force_terminal=force_terminal(), theme=theme(), stderr=True, **kwargs
            )
        case IO():
            return Console(theme=theme(), file=file, **kwargs)
        case file:
            if "width" not in kwargs:
                kwargs["width"] = 128
            return Console(theme=theme(), file=path.as_path(file).open("w"), **kwargs)


def force_terminal() -> bool | None:
    """...

    References:
        1. <https://force-color.org/>
        2. <https://no-color.org/>
    """
    if env.bool("FORCE_COLOR", None):
        return True
    if env.bool("NO_COLOR", None):
        return False
    if env.bool("GITHUB_ACTIONS", None):
        return True
    return None



from pathlib import Path
from typing import Any, override

import autoregistry

from liblaf.grapes.typed import PathLike

from ._abc import Serde
from ._json import json
from ._toml import toml
from ._yaml import yaml

SERIALIZERS = autoregistry.Registry()
SERIALIZERS["json"] = json
SERIALIZERS["toml"] = toml
SERIALIZERS["yaml"] = yaml
SERIALIZERS["yml"] = yaml


class Auto(Serde):
    @override
    def load(self, fpath: PathLike, *, ext: str | None = None, **kwargs) -> Any:
        serde: Serde = self.get_serde(fpath, ext=ext)
        return serde.load(fpath, **kwargs)

    @override
    def save(
        self, path: PathLike, obj: Any, *, ext: str | None = None, **kwargs
    ) -> None:
        serde: Serde = self.get_serde(path, ext=ext)
        serde.save(path, obj, **kwargs)

    def get_serde(self, path: PathLike, *, ext: str | None = None) -> Serde:
        if ext is None:
            path: Path = Path(path)
            ext = path.suffix.lstrip(".")
        return SERIALIZERS[ext]  # pyright: ignore[reportReturnType]


auto = Auto()


def load(path: PathLike, *, ext: str | None = None, **kwargs) -> Any:
    return auto.load(path, ext=ext, **kwargs)


def save(path: PathLike, obj: Any, *, ext: str | None = None, **kwargs) -> None:
    auto.save(path, obj, ext=ext, **kwargs)



import contextlib
import enum
import types
from collections.abc import Callable, Iterable, Sequence
from typing import Self, overload

import attrs

from liblaf.grapes.timing import callback

from ._base import Callback, TimerRecords
from ._function import TimedFunction
from ._iterable import TimedIterable


class TimerMode(enum.StrEnum):
    CONTEXT_MANAGER = "context-manager"
    FUNCTION = "function"
    ITERABLE = "iterable"
    INLINE = "inline"


@attrs.define
class Timer(
    contextlib.AbstractAsyncContextManager,
    contextlib.AbstractContextManager,
    TimerRecords,
):
    _mode: TimerMode | None = None

    async def __aenter__(self) -> Self:
        self.mode = TimerMode.CONTEXT_MANAGER
        self.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        self.stop()

    @overload
    def __call__[**P, T](self, func: Callable[P, T], /) -> TimedFunction[P, T]: ...
    @overload
    def __call__[T](
        self, iterable: Iterable[T], /, *, total: int | None = None
    ) -> TimedIterable[T]: ...
    def __call__(
        self, func_or_iterable: Callable | Iterable, /, *, total: int | None = None
    ) -> TimedFunction | TimedIterable:
        if callable(func_or_iterable):
            return TimedFunction(
                func_or_iterable,
                name=self.name,
                timers=self.timers,
                callback_start=self.callback_start,
                callback_stop=self.callback_stop,
                callback_finish=self.callback_finish,
            )
        # if isinstance(func_or_iterable, Iterable):
        return TimedIterable(
            func_or_iterable,
            name=self.name,
            timers=self.timers,
            total=total,
            callback_start=self.callback_start,
            callback_stop=self.callback_stop,
            callback_finish=self.callback_finish,
        )

    def __enter__(self) -> Self:
        self.mode = TimerMode.CONTEXT_MANAGER
        self.start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        self.stop()

    @property
    def mode(self) -> TimerMode | None:
        return self._mode

    @mode.setter
    def mode(self, value: TimerMode) -> None:
        if self._mode == value:
            return
        self._mode = value
        match value:
            case TimerMode.CONTEXT_MANAGER:
                if self.name is None:
                    self.name = "Block"
                if self.callback_stop is None:
                    self.callback_stop = callback.log_record(depth=3)
            case TimerMode.INLINE:
                if self.callback_stop is None:
                    self.callback_stop = callback.log_record(depth=2)
            case _:
                pass

    def start(self) -> None:
        self.mode = TimerMode.INLINE
        super().start()


@overload
def timer[T](
    iterable: Iterable[T],
    /,
    *,
    name: str = "Iterable",
    timers: Sequence[str] = ("perf",),
    total: int | None = None,
    callback_start: Callback | None = None,
    callback_stop: Callback | None = None,
    callback_finish: Callback | None = None,
) -> TimedIterable[T]: ...
@overload
def timer[**P, T](
    func: Callable[P, T],
    /,
    *,
    name: str | None = None,
    timers: Sequence[str] = ("perf",),
    callback_start: Callback | None = None,
    callback_stop: Callback | None = None,
    callback_finish: Callback | None = None,
) -> TimedFunction[P, T]: ...
@overload
def timer(
    *,
    name: str | None = None,
    timers: Sequence[str] = ("perf",),
    callback_start: Callback | None = None,
    callback_stop: Callback | None = None,
    callback_finish: Callback | None = None,
) -> Timer: ...
def timer(
    func_or_iterable: Callable | Iterable | None = None,
    /,
    *,
    name: str | None = None,
    timers: Sequence[str] = ("perf",),
    total: int | None = None,
    callback_start: Callback | None = None,
    callback_stop: Callback | None = None,
    callback_finish: Callback | None = None,
) -> TimedFunction | TimedIterable | Timer:
    timer = Timer(
        name=name,
        timers=timers,
        callback_start=callback_start,
        callback_stop=callback_stop,
        callback_finish=callback_finish,
    )
    if func_or_iterable is None:
        return timer
    if callable(func_or_iterable):
        return timer(func_or_iterable)
    # if isinstance(func_or_iterable, Iterable):
    return timer(func_or_iterable, total=total)



from collections.abc import Generator, Iterable, Sequence

from rich.progress import Progress

from liblaf.grapes import timing

from ._progress import progress


def track[T](
    iterable: Iterable[T],
    *,
    description: str = "Progress",
    timers: bool | Sequence[timing.TimerName | str] = ["perf"],
    total: float | None = None,
    callback_start: timing.Callback | None = None,
    callback_stop: timing.Callback | None = None,
    callback_finally: timing.Callback | None = None,
) -> Generator[T]:
    if timers is True:
        timers = ["perf"]
    if total is None:
        total = try_len(iterable)
    prog: Progress = progress(total_is_unknown=total is None)
    if timers:
        if callback_stop is None:
            callback_stop = timing.callback.log_record(depth=5)
        if callback_finally is None:
            callback_finally = timing.callback.log_summary(depth=5)
        iterable: timing.TimedIterable[T] = timing.timer(
            iterable,
            name=description,
            timers=timers,
            total=int(total) if total is not None else None,
            callback_start=callback_start,
            callback_stop=callback_stop,
            callback_finish=callback_finally,
        )
        with prog:
            yield from prog.track(iterable, total=total, description=description)
    else:
        with prog:
            yield from prog.track(iterable, total=total, description=description)


def try_len(iterable: Iterable) -> int | None:
    try:
        return len(iterable)  # pyright: ignore[reportArgumentType]
    except TypeError:
        return None



from collections.abc import Generator


def strip_comments(
    text: str,
    comments: str = "#",
    *,
    strip: bool = True,
    strip_empty_lines: bool = True,
) -> Generator[str]:
    """Remove comments and optionally strip whitespace and empty lines from the given text.

    Args:
        text: The input text from which to remove comments.
        comments: The comment delimiter.
        strip: Whether to strip leading and trailing whitespace from each line.
        strip_empty_lines: Whether to remove empty lines from the output.

    Yields:
        Lines of text with comments removed and optionally stripped of whitespace and empty lines.
    """
    for raw_line in text.splitlines():
        line: str = raw_line.split(comments, 1)[0]
        if strip:
            line = line.strip()
        if strip_empty_lines and not line:
            continue
        yield line



from collections.abc import Buffer
from typing import Any, override

import msgspec

from ._abc import Serde


class JSON(Serde):
    @override
    def decode(self, buf: Buffer | str, **kwargs) -> Any:
        return msgspec.json.decode(buf, **kwargs)

    @override
    def encode(self, obj: Any, **kwargs) -> bytes:
        return msgspec.json.encode(obj, **kwargs)


json = JSON()



from collections.abc import Callable, Hashable

import loguru


def default_transform(record: "loguru.Record") -> Hashable:
    return (
        record["file"].path,
        record["function"],
        record["level"].no,
        record["line"],
        record["message"],
        record["module"],
        record["name"],
    )


def filter_once(
    as_hashable: "Callable[[loguru.Record], Hashable]" = default_transform,
) -> "loguru.FilterFunction":
    history: set[Hashable] = set()

    def filter_(record: "loguru.Record") -> bool:
        if not record["extra"].get("once", False):
            return True
        hashable: Hashable = as_hashable(record)
        if hashable in history:
            return False
        history.add(hashable)
        return True

    return filter_



from collections.abc import Sequence
from typing import Unpack

import loguru
from rich.console import Console

from liblaf.grapes import pretty
from liblaf.grapes.logging.filters import make_filter
from liblaf.grapes.logging.sink import (
    LevelColumn,
    LocationColumn,
    LoguruRichHandler,
    MessageColumn,
    RichLoggingColumn,
    TimeColumn,
)


def rich_handler(
    console: Console | None = None,
    columns: Sequence[RichLoggingColumn] | None = None,
    *,
    enable_link: bool = True,
    **kwargs: Unpack["loguru.BasicHandlerConfig"],
) -> "loguru.BasicHandlerConfig":
    if console is None:
        console = pretty.get_console("stderr")
    if columns is None:
        columns = [
            TimeColumn(),
            LevelColumn(),
            LocationColumn(enable_link=enable_link),
            MessageColumn(),
        ]
    kwargs["sink"] = LoguruRichHandler(console=console, columns=columns)
    kwargs["format"] = ""
    kwargs["filter"] = make_filter(kwargs.get("filter"))
    return kwargs



from pathlib import Path
from typing import Unpack

import loguru
from environs import env

from liblaf.grapes.logging.filters import make_filter


def jsonl_handler(
    **kwargs: Unpack["loguru.FileHandlerConfig"],
) -> "loguru.FileHandlerConfig":
    kwargs.setdefault("sink", env.path("LOGGING_JSONL", Path("run.log.jsonl")))
    kwargs["filter"] = make_filter(kwargs.get("filter"))
    kwargs.setdefault("serialize", True)
    kwargs.setdefault("mode", "w")
    return kwargs



import bisect
import functools
from collections.abc import Callable, Mapping
from typing import Any, Literal, NoReturn

import attrs
from rich.text import Text

from liblaf.grapes import pretty


@attrs.frozen
class Function:
    condition: Callable[..., bool]
    function: Callable
    precedence: int = 0


class NotFoundLookupError(LookupError):
    func: Callable
    args: tuple
    kwargs: Mapping

    def __init__(self, func: Callable, args: tuple, kwargs: Mapping) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        return f"{self.pretty_call.plain} could not be resolved."

    @property
    def pretty_call(self) -> Text:
        return pretty.call(self.func, self.args, self.kwargs)


def _always_true(*args, **kwargs) -> Literal[True]:  # noqa: ARG001
    return True


def _fallback(func: Callable) -> Callable[..., NoReturn]:
    def fallback(*args, **kwargs) -> NoReturn:
        raise NotFoundLookupError(func, args, kwargs)

    return fallback


class ConditionalDispatcher:
    functions: list[Function]
    fallback: Callable

    def __init__(self) -> None:
        self.functions = []

    def __call__(self, *args, **kwargs) -> Any:
        for func in self.functions:
            try:
                if func.condition(*args, **kwargs):
                    return func.function(*args, **kwargs)
            except TypeError:
                continue
        return self.fallback(*args, **kwargs)

    def final(self, /, *, fallback: bool = False) -> Callable:
        def decorator[**P, T](func: Callable[P, T]) -> Callable[P, T]:
            if fallback:
                self.fallback = func
            else:
                self.fallback = _fallback(func)
            functools.update_wrapper(self, func)
            return self

        return decorator

    def register(
        self, condition: Callable[..., bool] = _always_true, *, precedence: int = 0
    ) -> Callable:
        def decorator[**P, T](func: Callable[P, T]) -> Callable[P, T]:
            function = Function(condition, func, precedence)
            bisect.insort(self.functions, function, key=lambda f: -f.precedence)
            return func

        return decorator



from loguru import logger

from liblaf.grapes import deps


def init_icecream() -> None:
    if not deps.has_module("icecream"):
        logger.warning(
            "`icecream` is not available. Skipping initialization of `icecream`."
        )
        return
    from icecream import ic

    ic.configureOutput(prefix="", outputFunction=icecream_output_function)


def icecream_output_function(s: str) -> None:
    logger.opt(depth=2).log("ICECREAM", s)



import contextlib
from collections.abc import Sequence

import loguru
from loguru import logger

DEFAULT_LEVELS: Sequence["loguru.LevelConfig"] = [
    {"name": "ICECREAM", "no": 15, "color": "<magenta><bold>", "icon": "🍦"}
]


def add_level(
    name: str, no: int, color: str | None = None, icon: str | None = None
) -> None:
    """Add a new logging level to the logger.

    Args:
        name: The name of the new logging level.
        no: The numeric value of the new logging level.
        color: The color associated with the new logging level.
        icon: The icon associated with the new logging level.
    """
    with contextlib.suppress(ValueError):
        logger.level(name, no, color=color, icon=icon)



import functools
import logging
import sys
import types
from collections.abc import Sequence
from typing import Unpack

import loguru
from environs import env
from loguru import logger

from ._intercept import setup_loguru_logging_intercept
from ._level import DEFAULT_LEVELS, add_level
from ._std import clear_handlers
from .handler import file_handler, jsonl_handler, rich_handler


def init_loguru(
    handlers: Sequence["loguru.HandlerConfig"] | None = None,
    levels: Sequence["loguru.LevelConfig"] | None = None,
    **kwargs: Unpack["loguru.BasicHandlerConfig"],
) -> None:
    traceback_install()
    if handlers is None:
        handlers: list[loguru.HandlerConfig] = [rich_handler(**kwargs)]
        if env.path("LOGGING_FILE", default=None):
            handlers.append(file_handler(**kwargs))  # pyright: ignore[reportArgumentType]
        if env.path("LOGGING_JSONL", default=None):
            handlers.append(jsonl_handler(**kwargs))  # pyright: ignore[reportArgumentType]
    logger.configure(handlers=handlers)
    for lvl in levels or DEFAULT_LEVELS:
        add_level(**lvl)
    setup_loguru_logging_intercept(kwargs.get("level", logging.NOTSET))
    clear_handlers()


def traceback_install(level: int | str = "CRITICAL", message: str = "") -> None:
    sys.excepthook = functools.partial(excepthook, level=level, message=message)


def excepthook(
    exc_type: type[BaseException],
    exc_value: BaseException,
    traceback: types.TracebackType,
    *,
    level: int | str = "CRITICAL",
    message: str = "",
) -> None:
    logger.opt(exception=(exc_type, exc_value, traceback)).log(level, message)



from pathlib import Path

from liblaf.grapes.typed import PathLike


def as_path(path: PathLike, *, expend_user: bool = True) -> Path:
    path = Path(path)
    if expend_user:
        path = path.expanduser()
    return path



from ._conditional_dispatcher import ConditionalDispatcher
from ._decorator import decorator_with_optional_arguments

__all__ = [
    "ConditionalDispatcher",
    "decorator_with_optional_arguments",
]



from rich.console import Console, RenderableType
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    Task,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.table import Column
from rich.text import Text

from liblaf.grapes import human as _human
from liblaf.grapes import pretty


class RateColumn(ProgressColumn):
    """RateColumn is a subclass of ProgressColumn that represents the rate of progress for a given task."""

    unit: str = "it"
    """The unit of measurement for the progress bar."""

    def __init__(self, unit: str = "it", table_column: Column | None = None) -> None:
        """.

        Args:
            unit: The unit of measurement for the progress bar.
            table_column: The table column associated with the progress bar.
        """
        super().__init__(table_column)
        self.unit = unit

    def render(self, task: Task) -> RenderableType:
        """Render the progress speed of a given task.

        Args:
            task: The task for which the speed is to be rendered.

        Returns:
            A text object representing the speed of the task.
        """
        if not task.speed:
            return Text(f"?{self.unit}/s", style="progress.data.speed")
        human: str = _human.human_throughout(task.speed, self.unit)
        return Text(human, style="progress.data.speed")


def progress(
    *columns: str | ProgressColumn,
    console: Console | None = None,
    total_is_unknown: bool = False,
) -> Progress:
    if not columns:
        columns: list[ProgressColumn] = [
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
        ]
        if total_is_unknown:
            columns += [
                MofNCompleteColumn(),
                "[",
                TimeElapsedColumn(),
                ",",
                RateColumn(),
                "]",
            ]
        else:
            columns += [
                TaskProgressColumn(),
                "[",
                TimeElapsedColumn(),
                "<",
                TimeRemainingColumn(),
                ",",
                RateColumn(),
                "]",
            ]
    console = console or pretty.get_console("stderr")
    progress = Progress(*columns, console=console)
    return progress



import datetime
import types
from collections.abc import Generator, Sequence
from typing import Protocol, override

import attrs
import loguru
from rich.console import Console, RenderableType
from rich.highlighter import Highlighter, ReprHighlighter
from rich.text import Text
from rich.traceback import Traceback

from liblaf.grapes import pretty


class RichLoggingColumn(Protocol):
    def render(self, record: "loguru.Record") -> RenderableType: ...


@attrs.define
class LoguruRichHandler:
    console: Console = attrs.field(factory=lambda: pretty.get_console("stderr"))
    columns: Sequence[RichLoggingColumn] = attrs.field(
        factory=lambda: [TimeColumn(), LevelColumn(), LocationColumn(), MessageColumn()]
    )

    def __call__(self, message: "loguru.Message") -> None:
        record: loguru.Record = message.record
        # TODO: `console.print()` is slow
        self.console.print(*self.render(record))
        if (excpetion := self.render_exception(record)) is not None:
            self.console.print(excpetion)

    def render(self, record: "loguru.Record") -> Generator[RenderableType]:
        for column in self.columns:
            yield column.render(record)

    def render_exception(self, record: "loguru.Record") -> RenderableType | None:
        exception: loguru.RecordException | None = record["exception"]
        if exception is None:
            return None
        exc_type: type[BaseException] | None
        exc_value: BaseException | None
        traceback: types.TracebackType | None
        exc_type, exc_value, traceback = exception
        if exc_type is None or exc_value is None:
            return None
        return Traceback.from_exception(
            exc_type=exc_type,
            exc_value=exc_value,
            traceback=traceback,
            width=self.console.width,
            code_width=self.console.width,
            show_locals=True,
        )


class TimeColumn(RichLoggingColumn):
    @override
    def render(self, record: "loguru.Record") -> RenderableType:
        elapsed: datetime.timedelta = record["elapsed"]
        hh: int
        mm: int
        ss: int
        mm, ss = divmod(int(elapsed.total_seconds()), 60)
        hh, mm = divmod(mm, 60)
        return Text(
            f"{hh:02d}:{mm:02d}:{ss:02d}.{elapsed.microseconds:06d}", style="log.time"
        )


@attrs.define
class LevelColumn(RichLoggingColumn):
    @override
    def render(self, record: "loguru.Record") -> RenderableType:
        level: str = record["level"].name
        return Text(f"{level:<8}", style=f"logging.level.{level.lower()}")


@attrs.define
class LocationColumn(RichLoggingColumn):
    enable_link: bool = attrs.field(default=True)

    @override
    def render(self, record: "loguru.Record") -> RenderableType:
        location: Text = pretty.location(
            name=record["name"],
            function=record["function"],
            line=record["line"],
            file=record["file"].path,
            enable_link=self.enable_link,
        )
        location.style = "log.path"
        return location


@attrs.define
class MessageColumn(RichLoggingColumn):
    highlighter: Highlighter = attrs.field(factory=ReprHighlighter)

    @override
    def render(self, record: "loguru.Record") -> RenderableType:
        if (rich := record["extra"].get("rich")) is not None:
            return rich
        message: RenderableType = record["message"].rstrip()
        if record["extra"].get("markup", True):
            message = Text.from_markup(message)
        if highlighter := record["extra"].get("highlighter", self.highlighter):
            message = highlighter(message)
        return message



import functools
from collections.abc import Callable

from liblaf.grapes.typed import (
    Decorator,
    DecoratorWithArguments,
    DecoratorWithOptionalArguments,
)


def decorator_with_optional_arguments[**Q](
    caller: DecoratorWithArguments[Q],
) -> DecoratorWithOptionalArguments[Q]:
    @functools.wraps(caller)
    def wrapper[**P, T](
        func: Callable[P, T] | None = None, /, *args, **kwargs
    ) -> Callable[P, T] | Decorator:
        if func is None:
            return caller(*args, **kwargs)
        return caller(*args, **kwargs)(func)

    return wrapper  # pyright: ignore[reportReturnType]



import functools
from collections.abc import Callable, Generator


def generator_to_list[**P, T](fn: Callable[P, Generator[T]]) -> Callable[P, list[T]]:
    @functools.wraps(fn)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> list[T]:
        return list(fn(*args, **kwargs))

    return wrapped



import collections
import statistics
import textwrap
from collections.abc import Callable, Generator, Mapping, Sequence
from typing import overload, override

import attrs
from loguru import logger

from liblaf.grapes import human

from ._time import get_time


@attrs.define
class SingleTimer:
    name: str | None = attrs.field(default=None, kw_only=True)
    timers: Sequence[str] = attrs.field(default=("perf",), kw_only=True)
    _time_start: dict[str, float] = attrs.field(init=False, factory=dict)
    _time_stop: dict[str, float] = attrs.field(init=False, factory=dict)

    @property
    def default_timer(self) -> str:
        return self.timers[0]

    @property
    def _current_record(self) -> Mapping[str, float]:
        return {timer: self.elapsed(timer) for timer in self.timers}

    def elapsed(self, timer: str | None = None) -> float:
        if timer is None:
            timer = self.default_timer
        if timer in self._time_stop:
            return self._time_stop[timer] - self._time_start[timer]
        return get_time(timer) - self._time_start[timer]

    def start(self) -> None:
        for timer in self.timers:
            self._time_start[timer] = get_time(timer)
        self._time_stop.clear()

    def stop(self) -> None:
        for timer in self.timers:
            self._time_stop[timer] = get_time(timer)


type Callback = Callable[["TimerRecords"], None]


@attrs.define
class TimerRecords(SingleTimer):
    callback_start: Callback | None = attrs.field(default=None, kw_only=True)
    callback_stop: Callback | None = attrs.field(default=None, kw_only=True)
    callback_finish: Callback | None = attrs.field(default=None, kw_only=True)
    _records: dict[str, list[float]] = attrs.field(
        init=False, factory=lambda: collections.defaultdict(list)
    )

    @overload
    def __getitem__(self, key: int) -> Mapping[str, float]: ...
    @overload
    def __getitem__(self, key: str) -> Sequence[float]: ...
    def __getitem__(self, key: int | str) -> Mapping[str, float] | Sequence[float]:
        if isinstance(key, int):
            return {k: v[key] for k, v in self._records.items()}
        if isinstance(key, str):
            return self._records[key]
        raise KeyError(key)

    def __len__(self) -> int:
        return len(self._records[self.default_timer])

    @property
    def columns(self) -> Sequence[str]:
        return self.timers

    @property
    def count(self) -> int:
        return self.n_rows

    @property
    def n_columns(self) -> int:
        return len(self.timers)

    @property
    def n_rows(self) -> int:
        return len(self.column())

    def column(self, timer: str | None = None) -> Sequence[float]:
        if timer is None:
            timer = self.default_timer
        return self._records[timer]

    def human_record(self, index: int = -1) -> str:
        return human_record(self.row(index), name=self.name)

    def human_summary(self) -> str:
        name: str = self.name or "Timer"
        header: str = f"{name} (total: {self.n_rows})"
        if self.n_rows == 0:
            return header
        body: str = ""
        for timer in self.columns:
            body += f"{timer} > "
            human_mean: str = human.human_duration_series(self.column(timer))
            human_median: str = human.human_duration(self.median(timer))
            body += f"mean: {human_mean}, median: {human_median}\n"
        body = body.strip()
        summary: str = header + "\n" + textwrap.indent(body, "    ")
        return summary

    def iter_columns(self) -> Generator[tuple[str, Sequence[float]]]:
        yield from self._records.items()

    def iter_rows(self) -> Generator[Mapping[str, float]]:
        for index in range(self.n_rows):
            yield self.row(index)

    def log_record(
        self, index: int = -1, depth: int = 1, level: int | str = "DEBUG"
    ) -> None:
        logger.opt(depth=depth).log(level, self.human_record(index=index))

    def log_summary(self, depth: int = 1, level: int | str = "INFO") -> None:
        logger.opt(depth=depth).log(level, self.human_summary())

    def row(self, index: int) -> Mapping[str, float]:
        return {timer: values[index] for timer, values in self._records.items()}

    # region statistics

    def max(self, timer: str | None = None) -> float:
        return max(self.column(timer))

    def mean(self, timer: str | None = None) -> float:
        return statistics.mean(self.column(timer))

    def median(self, timer: str | None = None) -> float:
        return statistics.median(self.column(timer))

    def min(self, timer: str | None = None) -> float:
        return min(self.column(timer))

    def std(self, timer: str | None = None) -> float:
        return statistics.stdev(self.column(timer))

    # endregion statistics

    def _append(
        self, seconds: Mapping[str, float] = {}, nanoseconds: Mapping[str, float] = {}
    ) -> None:
        for key, value in seconds.items():
            self._records[key].append(value)
        for key, value in nanoseconds.items():
            self._records[key].append(value * 1e-9)

    @override
    def start(self) -> None:
        if callable(self.callback_start):
            self.callback_start(self)
        super().start()

    @override
    def stop(self) -> None:
        super().stop()
        self._append(seconds=self._current_record)
        if callable(self.callback_stop):
            self.callback_stop(self)

    def finish(self) -> None:
        if callable(self.callback_finish):
            self.callback_finish(self)


def human_record(record: Mapping[str, float], name: str | None = None) -> str:
    name = name or "Timer"
    text: str = f"{name} > "
    items: list[str] = []
    for timer, value in record.items():
        human_duration: str = human.human_duration(value)
        items.append(f"{timer}: {human_duration}")
    text += ", ".join(items)
    return text


def human_summary(records: Mapping[str, Sequence[float]], name: str | None) -> str:
    name = name or "Timer"
    header: str = f"{name} (total: {len(records)})"
    if len(records) == 0:
        return header
    body: str = ""
    for timer, values in records.items():
        body += f"{timer} > "
        human_mean: str = human.human_duration_series(values)
        human_median: str = human.human_duration(statistics.median(values))
        body += f"mean: {human_mean}, median: {human_median}\n"
    body = body.strip()
    summary: str = header + "\n" + textwrap.indent(body, "    ")
    return summary



from collections.abc import Sequence
from typing import Any

from liblaf.grapes.typed import ClassInfo

from ._as_iterable import as_iterable


def as_sequence(obj: Any, base_type: ClassInfo | None = (str, bytes)) -> Sequence:
    return tuple(as_iterable(obj, base_type))



from . import (
    const,
    environ,
    functools,
    git,
    human,
    itertools,
    logging,
    path,
    pretty,
    serde,
    text,
    timing,
    tqdm,
    typed,
)
from ._version import __version__, __version_tuple__, version, version_tuple
from .const import nop
from .deps import has_module, optional_imports
from .environ import init_env
from .functools import ConditionalDispatcher, decorator_with_optional_arguments
from .human import (
    human_count,
    human_duration,
    human_duration_series,
    human_duration_unit_precision,
    human_duration_with_variance,
    human_throughout,
)
from .itertools import IterableWrapper, as_iterable, as_sequence, generator_to_list
from .logging import (
    critical_once,
    debug_once,
    error_once,
    exception_once,
    info_once,
    init_logging,
    log_once,
    success_once,
    trace_once,
    warning_once,
)
from .path import as_path, project_root, resolve_project_path
from .pretty import caller_location, get_console
from .serde import json, load, load_pydantic, save, save_pydantic, toml, yaml
from .text import strip_comments
from .timing import TimedFunction, TimedIterable, TimerName, get_time, timer
from .tqdm import progress, track

__all__ = [
    "ConditionalDispatcher",
    "IterableWrapper",
    "TimedFunction",
    "TimedIterable",
    "TimerName",
    "__version__",
    "__version_tuple__",
    "as_iterable",
    "as_path",
    "as_sequence",
    "caller_location",
    "caller_location",
    "const",
    "critical_once",
    "debug_once",
    "decorator_with_optional_arguments",
    "environ",
    "error_once",
    "exception_once",
    "functools",
    "generator_to_list",
    "get_console",
    "get_time",
    "git",
    "has_module",
    "human",
    "human_count",
    "human_duration",
    "human_duration_series",
    "human_duration_unit_precision",
    "human_duration_with_variance",
    "human_throughout",
    "info_once",
    "init_env",
    "init_logging",
    "itertools",
    "json",
    "load",
    "load_pydantic",
    "log_once",
    "logging",
    "nop",
    "optional_imports",
    "path",
    "pretty",
    "progress",
    "project_root",
    "resolve_project_path",
    "save",
    "save_pydantic",
    "save_pydantic",
    "serde",
    "strip_comments",
    "success_once",
    "text",
    "timer",
    "timing",
    "toml",
    "tqdm",
    "trace_once",
    "track",
    "typed",
    "version",
    "version_tuple",
    "warning_once",
    "yaml",
]



import functools

from loguru import logger

from ._base import Callback, TimerRecords


def log_record(
    *,
    depth: int = 1,
    elapsed_threshold: float = 1e-3,
    index: int = -1,
    level: int | str = "DEBUG",
) -> Callback:
    return functools.partial(
        _log_record,
        depth=depth,
        elapsed_threshold=elapsed_threshold,
        index=index,
        level=level,
    )


def log_summary(*, depth: int = 1, level: int | str = "INFO") -> Callback:
    return functools.partial(_log_summary, depth=depth, level=level)


def _log_record(
    timer: TimerRecords,
    *,
    depth: int = 1,
    elapsed_threshold: float = 1e-3,
    index: int = -1,
    level: int | str = "DEBUG",
) -> None:
    if timer.elapsed() < elapsed_threshold:
        return
    logger.opt(depth=depth).log(level, timer.human_record(index=index))


def _log_summary(
    timer: TimerRecords, *, depth: int = 1, level: int | str = "INFO"
) -> None:
    logger.opt(depth=depth).log(level, timer.human_summary())


__all__ = ["log_record", "log_summary"]



import rich.pretty
import rich.traceback
from typing_extensions import deprecated

from liblaf.grapes import pretty


@deprecated("This function is deprecated and will be removed in a future version.")
def init_rich(*, show_locals: bool = True) -> None:
    """Initialize rich logging for pretty printing and tracebacks.

    This function sets up rich's pretty printing and traceback handling
    for the logging console.

    Args:
        show_locals: If True, local variables will be shown in tracebacks.
    """
    rich.pretty.install(console=pretty.get_console("stderr"))
    rich.traceback.install(
        console=pretty.get_console("stderr"), show_locals=show_locals
    )



from collections.abc import Callable
from typing import Protocol, overload


class Decorator(Protocol):
    def __call__[**P, T](self, func: Callable[P, T], /) -> Callable[P, T]: ...


class DecoratorWithArguments[**P](Protocol):
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Decorator: ...


class DecoratorWithOptionalArguments[**Q](Protocol):
    @overload
    def __call__[**P, T](self, func: Callable[P, T], /) -> Callable[P, T]: ...
    @overload
    def __call__(self, *args: Q.args, **kwargs: Q.kwargs) -> Decorator: ...



from . import callback
from ._base import Callback, TimerRecords
from ._function import TimedFunction
from ._iterable import TimedIterable
from ._time import TimerName, get_time
from ._timer import Timer, timer

__all__ = [
    "Callback",
    "TimedFunction",
    "TimedIterable",
    "Timer",
    "TimerName",
    "TimerRecords",
    "callback",
    "get_time",
    "timer",
]



from pathlib import Path

from liblaf.grapes import deps
from liblaf.grapes.typed import PathLike

with deps.optional_imports(extra="git"):
    import git
    import git.exc


def root(
    path: PathLike | None = None, *, search_parent_directories: bool = True
) -> Path:
    repo = git.Repo(path=path, search_parent_directories=search_parent_directories)
    return Path(repo.working_dir)


def root_safe(
    path: PathLike | None = None, *, search_parent_directories: bool = True
) -> Path:
    try:
        return root(path=path, search_parent_directories=search_parent_directories)
    except git.exc.InvalidGitRepositoryError:
        return Path()



from collections.abc import Iterable, Iterator, Sequence

from . import callback
from ._base import Callback, TimerRecords


class TimedIterable[T]:
    timing: TimerRecords
    _iterable: Iterable[T]
    _total: int | None = None

    def __init__(
        self,
        iterable: Iterable[T],
        /,
        total: int | None = None,
        name: str | None = None,
        timers: Sequence[str] = ("perf",),
        callback_start: Callback | None = None,
        callback_stop: Callback | None = None,
        callback_finish: Callback | None = None,
    ) -> None:
        if name is None:
            name = "Iterable"
        if callback_stop is None:
            callback_stop = callback.log_record(depth=3)
        if callback_finish is None:
            callback_finish = callback.log_summary(depth=3)
        self.timing = TimerRecords(
            name=name,
            timers=timers,
            callback_start=callback_start,
            callback_stop=callback_stop,
            callback_finish=callback_finish,
        )
        self._iterable = iterable
        self._total = total

    def __contains__(self, x: object, /) -> bool:
        return x in self._iterable  # pyright: ignore[reportOperatorIssue]

    def __len__(self) -> int:
        if self._total is None:
            return len(self._iterable)  # pyright: ignore[reportArgumentType]
        return self._total

    def __iter__(self) -> Iterator[T]:
        for item in self._iterable:
            self.timing.start()
            yield item
            self.timing.stop()
        self.timing.finish()



import loguru

from ._as_filter_func import as_filter_func
from .typed import Filter


def filter_all(*filters: Filter | None) -> "loguru.FilterFunction":
    filters: list[loguru.FilterFunction] = [
        fn for f in filters if (fn := as_filter_func(f)) is not None
    ]

    def filter_(record: "loguru.Record") -> bool:
        return all(fn(record) for fn in filters)

    return filter_


def filter_any(*filters: Filter | None) -> "loguru.FilterFunction":
    filters: list[loguru.FilterFunction] = [
        fn for f in filters if (fn := as_filter_func(f)) is not None
    ]

    def filter_(record: "loguru.Record") -> bool:
        return any(fn(record) for fn in filters)

    return filter_
